import java.util.Random;
import java.util.Vector;

/** MotorInferencias.java
   Implementa el motor de inferencias de un sistema experto que
   juega al conecta 4, es decir, el sistema que se va encargando de
   seleccionar y aplicar las posibles operaciones sobre la base de
   hechos.
   @author Alberto Alonso Ruibal
*/

class MotorInferencias {
	static int iVALOR_VICTORIA = 9999;
	/** Limite de la profundidad de exploracion */
	static int iLIMITE_PROFUNDIDAD = 6;
	static int iVALORES_CASILLAS[][] =
	{{1,2,2,2,2,1},
	 {2,3,3,3,3,2},
	 {3,4,5,5,4,3},
	 {3,4,5,5,4,3},
	 {3,4,5,5,4,3},
	 {2,3,3,3,3,2},
	 {1,2,2,2,2,1}};

	BaseHechos oBaseHechos;
	BaseConocimientos oBaseConocimientos;

	int iValorObtenido;
    Operacion oMejorOperacion;

	/**
	   Construye un motor de inferencias con su base de hechos
	   y su base de conocimiento.
	*/
	MotorInferencias() {
		oBaseHechos = new BaseHechos();
		oBaseConocimientos = new BaseConocimientos(oBaseHechos);
	}

	/**
	   Hace un movimiento sobre el tablero. Si le pasamos en comprobar
	   false, ya no comprueba la correccion del movimiento (para
	   ganar velocidad)
	 */
	int aplicaOperacion(Operacion oOperacion, boolean bComprobar) {
		int i;
		if (oOperacion == null) return -1;

		// Comprueba que se pueda realizar el movimiento
		if (bComprobar)
			if (!oBaseConocimientos.operacionValida(oOperacion))
				return -1;

		for (i = 0; i < 6; i++)
			if (oBaseHechos.iTablero[oOperacion.intValue()][i] == 0) {
				oBaseHechos.iTablero[oOperacion.intValue()][i] =
					(oBaseHechos.bTurno ?
					 BaseHechos.iFICHA_BLANCA :
					 BaseHechos.iFICHA_NEGRA);
				oBaseHechos.bTurno = ! oBaseHechos.bTurno;
				return 0;
			}
		return 0;
	}


	/**
	   Hace una estimacion heurstica del valor de la posicion.
	   Si el valor que devuelve es positivo la posicion es favorable
	   a las blancas y cuanto ms grande sea el nmero, ms favorable
	   ser la posicion.
	 */
	int valoraEstado() {
		int i, j, k;
		int iValor = 0;
		int iNumeroFichas = 0; // Lleva la cuenta del numero de fichas
		Random oRandom = new Random(System.currentTimeMillis());

		for (i = 0; i < 7; i++)
			for (j = 0; j < 6; j++)
				if ((k = oBaseHechos.iTablero[i][j]) != 0) {
					iNumeroFichas ++;
					iValor += (k == BaseHechos.iFICHA_BLANCA ?
							   iVALORES_CASILLAS[i][j] :
							   -iVALORES_CASILLAS[i][j]);
				}
		// Las victorias mas valiosas son las quue se consiguen
		// en un menor numero de movimientos (con una menos cantidad
		// de fichas)
		if ((k = oBaseConocimientos.hay4enRaya()) != 0)
			iValor = (k == BaseHechos.iFICHA_BLANCA ?
					(iVALOR_VICTORIA - iNumeroFichas) :
					-(iVALOR_VICTORIA - iNumeroFichas));

		// Aplicamos algo de aleatoriedad...
		iValor += ((oRandom.nextInt() & 0xf0)>>4) - 2;
		return iValor;
	}

	/**
	   Se llama "recursivamente" (crea un nuevo motor de inferencias en
	   cada llamada) para implementar una busqueda en profundidad de la
	   mejor operacin a aplicar.
	   Devuelve -1 si no se puede realizar ningun movimiento.
	   Da un valor a las variables iValorObtenido y oMejorOperacion.
	   El parametro iLimite es usado para la simplificacion de Shannon, 
	   si en en esta posicion hay un valor mas favorable a los que juegan
	   ahora que el mejor valor en la posicion anterior a esta (=iLimite),
	   ya no se prosigue con la exploracion porque sabemos que al volver
	   atrs, en el anterior nodo no se va a tomar este camino.
	*/
	int seleccionaOperacion(int iProfundidad, int iLimite) {
		int i;
		Vector oOperaciones;
		Operacion oOperacion;
		MotorInferencias oMotorInferencias;
		oMotorInferencias = new MotorInferencias();

		oMejorOperacion = null;
		iValorObtenido = (oBaseHechos.bTurno ?
						  -iVALOR_VICTORIA : iVALOR_VICTORIA);

		// Mira cundo retrocede en la exploracion del rbol
		if ((oBaseConocimientos.hay4enRaya() != 0) ||
			(iProfundidad >= iLIMITE_PROFUNDIDAD )) {
			iValorObtenido = valoraEstado();
			oMejorOperacion = null;
			return -1;
		}

		// Cogemos la lista de movimientos posibles y la vamos
		// recorriendo
		oOperaciones = oBaseConocimientos.listaOperaciones();
		for (i = 0; i < oOperaciones.size(); i++) {
			oOperacion = (Operacion) oOperaciones.elementAt(i);
			oMotorInferencias.oBaseHechos.copia(oBaseHechos);
			oMotorInferencias.aplicaOperacion(oOperacion, false);

			oMotorInferencias.seleccionaOperacion(iProfundidad + 1,
												  iValorObtenido);

			if (oBaseHechos.bTurno) { // Si juegan blancas...
				if (iValorObtenido < oMotorInferencias.iValorObtenido) {
					iValorObtenido = oMotorInferencias.iValorObtenido;
					oMejorOperacion = oOperacion;
					// Principio de simplificacion de Shannon
					if (iValorObtenido > iLimite) return 0;
				}
			} else { // ... y si juegan negras...
				if (iValorObtenido > oMotorInferencias.iValorObtenido) {
					iValorObtenido = oMotorInferencias.iValorObtenido;
					oMejorOperacion = oOperacion;
					// Principio de simplificacion de Shannon
					if (iValorObtenido < iLimite) return 0;
				}
			}
		}
		return 0;
	}

	/**
	   Busca el mejor movimiento y mueve
	*/
	public int juega() {
		int iLimite = (oBaseHechos.bTurno ?
					   iVALOR_VICTORIA : -iVALOR_VICTORIA);
		seleccionaOperacion(0, iLimite);
		return aplicaOperacion(oMejorOperacion, true);
	}
}

